﻿// Dependencies
//#region Dependencies

// External (Common) dependencies
var fs = require('fs');
var cp = require('child_process');
var moment = require('moment');

// External (Device) dependencies
var Parser = require('expr-eval').Parser;// Library to manage parsing and evaluation of formulas
var mathparser = new Parser();

// Local dependencies
var Logger;// Used for logging
var SocketIo;// Used for messaging
var NetSnmp;// Used for SNMP messaging - base library
var SmtpManager;// Used for SMTP messaging
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions

//#endregion Dependencies

// Constants
//#region Constants

// Constants (Common)
//#region Constants (Common)
//  Device
var DEVICE_MODEL = "./DoorSensor.json"; //  Device model defines the device incl. general properties, parameters and I/O 

//  Event Log
//#region Event Log
var EVENTS_LENGTH = 10;
var ThresholdValues = ["NORMAL", "BELOW", "ABOVE"];
//#endregion Event Log
//#endregion Constants (Common)

// Constants (Device)
//#region Constants (Device)
// Holding States
var HOLDSTATE_OFF = "red";// Active
var HOLDSTATE_ACTIVE = "green";// Active
var HOLDSTATE_HOLDING = "#8F8F8F"// Changing
var HOLDSTATE_UNKNOWN = "#EB6F21";// Orangy

// Analog Inputs
//#region Analog Inputs
//  Analog input offset will be determined by selected port #
var ANALOG_INPUTS_LOCATION = '/sys/class/hwmon/hwmon0';

//  Analog input raw to voltage linear equation
var ANALOG_INPUTS_RAWVALUE_OFFSET = -0.075;
var ANALOG_INPUTS_RAWVALUE_SLOPE = 0.005;
var ANALOG_INPUTS_CHANGE_FILTER = 4;// Look for 20mV change
var ANALOG_INPUTS_TTL_ONOFF = 2.00;// Amount of voltage to switch to ON
var ANALOG_INPUTS_FIRST_READ = true;// Track the first read

// Analog threshold states
var ANALOG_INPUTS_ILLEGAL = -1;
var ANALOG_INPUTS_IN_RANGE = 0;
var ANALOG_INPUTS_BELOW_THRESHOLD = 1;
var ANALOG_INPUTS_ABOVE_THRESHOLD = 2;
var AnalogThresholdValues = ["NORMAL", "BELOW", "ABOVE"];
//#endregion Analog Inputs

// Power Control
//#region Power Control
//  Power control offset will be determined by selected port #
var POWER_CONTROL_LOCATION = '/sys/class/gpio/gpio';
var POWER_CONTROL_BASEGPIO = 496;
//#endregion Power Control

//#endregion Constants (Device)

//#endregion Constants

// Device Class: Variables, Instantiation, Methods and Constructor
//#region Device Class: Variables, Instantiation, Methods and Constructor
function Device(port, logger, io, smtpManager, netSnmp) {

    // Class Variables
    //#region Class Variables

    // Class Variables (Common)
    //#region Class Variables (Common)
    var EventLog = [];
    var EnabledFlag = false;
    var SendEventsFlag = false;
    var DeviceModel;
    var ThisDeviceObject = this;
    //#endregion Class Variables (Common)

    // Class Variables (Device-specific)
    //#region Class Variables (Device-specific)
    // Digital input
    var DigitalCycleControl;// Manages the setInterval timer for digital I/O
    var DigitalInputsCycle;// Cycle speed
    var DigitalInterimUpdate;
    var DigitalSendInterimUpdate = false;
    var DigitalLastTimeBeforeInterimUpdate = moment();
    var DigitalCurrentTimeBeforeInterimUpdate;
    //#endregion Class Variables (Device-specific)

    //#endregion Class Variables

    // Class Instantiation
    //#region Class Instantiation

    // Local Dependencies
    //#region Local Dependencies
    Logger = logger;
    SocketIo = io;
    SmtpManager = smtpManager;
    NetSnmp = netSnmp;
    //#endregion Local Dependencies

    try {
        // Load a private model for this instance of the device
        //  Remove model from cache to avoid overwrites
        var cache = require.resolve(DEVICE_MODEL);
        if (cache) delete require.cache[cache];
        //  Load model
        var DeviceModel = require(DEVICE_MODEL);
        //  Assign statics to model
        DeviceModel.port = port;
        DeviceModel.uuid = parseInt(moment().format("x"));
        Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: DeviceModel=" + JSON.stringify(DeviceModel));
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Device uuid=" + DeviceModel.uuid + " -----------------------");
    } catch (error) {
        Logger.error("Failed constructing device on port=" + port + " error=" + error);
    }
    //#endregion Class Instantiation

    // Class Methods
    //#region Class Methods

    //#region Class Methods (Common)

    //#region Class Methods (Common - Basic)
    this.setPort = function (port) {
        DeviceModel.port = port;
    }

    this.getPort = function () {
        return DeviceModel.port;
    }

    this.getDeviceId = function () {
        return DeviceModel.id;
    }

    this.getDeviceModel = function () {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceModel: Device uuid=" + DeviceModel.uuid);
        return DeviceModel;
    }

    this.enable = function (enabled) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Enabled=" + (enabled ? "YES" : "NO"));
        if (enabled == EnabledFlag)
            return;// Don't do anything if we're already in the proper state
        else
            EnabledFlag = enabled;// Remember new setting

        if (enabled) {
            //#region Commission Device
            var power;
            if (DeviceModel.io.powerControl.enabled) power = 1; else power = 0;
            fs.truncate(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", 0, function () {
                fs.writeFile(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", power, function (err) {
                    if (err) {
                        return Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Power Set Error (Write)=" + JSON.stringify(err));
                    }
                    Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Port=" + DeviceModel.port + " - '" + POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value" + "' Power=" + power);
                });
            });
            //#endregion Commission Device

            //#region Start Device Processing
            this.startProcessing();
            //#endregion Start Device Processing
        }

        else {
            //#region Stop Device Processing
            this.stopProcessing();
            //#endregion Stop Device Processing

            //#region Decommission Device
            var power;
            if (DeviceModel.io.powerControl.enabled) power = 0; else power = 1;
            fs.truncate(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", 0, function () {
                fs.writeFile(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", power, function (err) {
                    if (err) {
                        return Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Power Set Error (Write)=" + JSON.stringify(err));
                    }
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Port=" + DeviceModel.port + " - '" + POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value" + "' Power=" + power);
                });
            });
            //#endregion Decommission Device
        }
    }

    this.sendEvents = function (sendevents) {
        SendEventsFlag = sendevents;
    }
    //#endregion Class Methods (Common - Basic)

    //#region Class Methods (Common - UpdateDevice)
    this.updateDevice = function (port, parameter, model, callback) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateDevice: Device uuid=" + DeviceModel.uuid);
        DeviceModel = model;
        // Force recalculation of parameters
        DeviceModel.parameters[0].RawValue = -1;// Since there is only one, no need to worry about other parameters
        callback(DeviceModel);
    }
    //#endregion Class Methods (Common - UpdateDevice)

    //#endregion Class Methods (Common)

    //#region Class Methods (Device Processing)

    //#region Class Methods (Device Processing - Start/Stop)
    //#region startProcessing
    this.startProcessing = function () {

        // Digital Inputs Cycle
        //DigitalInputsCycle = 5000;// Temporary
        DigitalInputsCycle = DeviceModel.io.digitalIo.cycleTime;
        DigitalInterimUpdate = DigitalInputsCycle * 20;
        DeviceModel.parameters[0].HoldTimer = 0;
        DeviceModel.parameters[0].Value = -1;
        DigitalCycleControl = setInterval(function () { ThisDeviceObject.processDigitalInput() }, DigitalInputsCycle);

        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".startProcessing: IO processing cycle started");
    }
    //#endregion startProcessing

    //#region stopProcessing
    this.stopProcessing = function () {

        //  Stop analog and digital processing
        try {
            if (DigitalCycleControl) clearInterval(DigitalCycleControl);
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopped IO processing cycle.");
        } catch (error) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopped IO processing cycle. Error=" + error);
        }
    }
    //#endregion stopProcessing

    //#endregion Class Methods (Device Processing - Start/Stop)

    //#region Class Methods (Device Processing - Device Information)
    //#region getDeviceInformation
    this.getDeviceInformation = function () {
        return DeviceModel.device;
    }
    //#endregion getDeviceInformation
    //#endregion Class Methods (Device Processing - Device Information)

    //#region Class Methods (Device Processing - Process I/O)

    //#region Digital I/O 
    //#region processDigitalInput
    //  Reads the value of the digital input (Door), based on hold time
    //  Updates UI and SNMP/SMTP based on state changes
    this.processDigitalInput = function () {

        var timeChanged = moment();
        var rawValue;

        // This is to allow interim updates in case the user has changed screens between udpates
        DigitalCurrentTimeBeforeInterimUpdate = timeChanged;
        if ((DigitalCurrentTimeBeforeInterimUpdate - DigitalLastTimeBeforeInterimUpdate) > DigitalInterimUpdate) {
            DigitalSendInterimUpdate = true;
            DigitalLastTimeBeforeInterimUpdate = DigitalCurrentTimeBeforeInterimUpdate;
        } else {
            DigitalSendInterimUpdate = false;
        }

        try {
            // Read Logic
            //  If the raw value has changed then we'll have to do stuff
            //#region Read Logic

            var rawValue = 0;
            var rangedValue = 0;
            var qualifierValue = false;
            var data = 0;
            var offset = parseInt(parseInt(DeviceModel.port) * 2 + parseInt(DeviceModel.io.analogIo.analog0.offset));
            try {
                Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Read: ACCESSIO] Reading from '" + ANALOG_INPUTS_LOCATION + "/in" + offset + "_input'");
                data = fs.readFileSync(ANALOG_INPUTS_LOCATION + "/in" + offset + "_input", 'utf8');
            } catch (err) {
                // This is to catch infrequent errors related to reading
                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Read: ACCESSIO] Error=" + err);
                return;
            }

            // Get rid of any offending extra characters in read
            rawValue = String(data).replace(/(\r\n|\n|\r|[ +])/img, "");
            //rawValue = parseInt(rawValue) + Math.floor(Math.random() * 1001);//*** RANDOM
            // Add the offset to the read value
            rawValue = Math.round(Number(rawValue) - Number(this.convertValueToRawValue(0)));
            rangedValue = this.convertRawValueToValue(rawValue);

            //#region Normally Closed vs Normally Open
            if (DeviceModel.parameters[0].Mode == "Normally Open") {
                if (rangedValue > ANALOG_INPUTS_TTL_ONOFF) {
                    qualifierValue = true;
                } else {
                    qualifierValue = false;
                }
            } else {
                if (rangedValue > ANALOG_INPUTS_TTL_ONOFF) {
                    qualifierValue = false;
                } else {
                    qualifierValue = true;
                }
            }
            //#endregion Normally Closed vs Normally Open
            Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: qualifierValue=" + qualifierValue);

            if ((DeviceModel.parameters[0].Value != qualifierValue) || (DeviceModel.parameters[0].HoldTime != 0)) {
                // Filter logic - we only consider changes if there was a:
                //  - Change in the digital value, or
                //  - A hold timer is being used
                //#region Filter logic                  
                //if (DeviceModel.parameters[0].HoldTime > 0) {
                //    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: HOLDTIMER=" + parseFloat(DeviceModel.parameters[0].HoldTime).toFixed(2) + "s] Previous=" + DeviceModel.parameters[0].Value + " Current=" + rawValue);
                //    Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: HOLDTIMER=" + parseFloat(DeviceModel.parameters[0].HoldTime).toFixed(2) + "s] Previous=" + DeviceModel.parameters[0].Value + " Current=" + rawValue);
                //} else {
                //    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: VALUECHANGE] Previous Value=" + DeviceModel.parameters[0].Value + " Current=" + qualifierValue);
                //    Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: VALUECHANGE] Previous Value=" + DeviceModel.parameters[0].Value + " Current=" + qualifierValue);
                //}
                //#endregion Filter logic

                // HoldTime logic
                //  - A non-zero value in the HoldTimer variable means that the HoldTimer is active
                //  - Start the timer if there was a change and the timer wasn't previously started
                //  - If the timer was previously started, then check if the 
                //#region HoldTime logic
                var HoldTimeValid = false;
                var HoldTimeChanged = 0;
                if (DeviceModel.parameters[0].HoldTime > 0) {
                    // HoldTime is used
                    //#region HoldTime is used
                    if (DeviceModel.parameters[0].Value != qualifierValue) {
                        // Value has changed ...
                        if (DeviceModel.parameters[0].HoldTimer === 0) {
                            // HoldTimer not started, so start it
                            //#region HoldTime off, start timer
                            DeviceModel.parameters[0].HoldTimer = timeChanged;// Start the timer
                            DeviceModel.parameters[0].HoldTimeState = HOLDSTATE_HOLDING;
                            this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters[0]);
                            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: STARTED] HoldTimer=0.00");
                            //#endregion HoldTime off, start timer
                        }
                        else {
                            HoldTimeChanged = (timeChanged - DeviceModel.parameters[0].HoldTimer) / 1000;
                            if (HoldTimeChanged >= DeviceModel.parameters[0].HoldTime) {
                                // HoldTime active and reached
                                //#region Holdtime active and reached
                                HoldTimeValid = true;
                                DeviceModel.parameters[0].HoldTimeState = (qualifierValue ? HOLDSTATE_ACTIVE : HOLDSTATE_OFF);
                                this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters[0]);
                                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: ACTIVE] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                                DeviceModel.parameters[0].HoldTimer = 0;// Set hold timer to 0
                                //#endregion Holdtime active and reached
                            } else {
                                // HoldTime active, but not reached
                                //#region HoldTime active, but not reached
                                DeviceModel.parameters[0].HoldTimeState = HOLDSTATE_HOLDING;
                                this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters[0]);
                                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: HOLDING] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                                //#endregion HoldTime active, but not reached
                            }
                        }
                    }
                    else {
                        // Value has either returned to the same value or hasn't changed, so reset the hold timer
                        //#region HoldTime is reset
                        if (DeviceModel.parameters[0].HoldTimer !== 0) {
                            // Returned to the same value
                            DeviceModel.parameters[0].HoldTimer = 0;
                            DeviceModel.parameters[0].HoldTimeState = (qualifierValue ? HOLDSTATE_ACTIVE : HOLDSTATE_OFF);
                            this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters[0]);
                            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: RESET] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                        }
                        //#endregion HoldTime is reset
                    }
                    //#endregion HoldTime is used
                } else {
                    // HoldTime is not used
                    //#region HoldTime is not used
                    HoldTimeValid = true;
                    DeviceModel.parameters[0].HoldTimeState = (qualifierValue ? HOLDSTATE_ACTIVE : HOLDSTATE_OFF);
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: NOT USED]");
                    //#endregion HoldTime is not used
                }
                //#endregion HoldTime logic

                // Publish logic
                //  - If the value has changed and the hold timer logic has been met, then change the value            
                //#region Publish logic
                if ((DeviceModel.parameters[0].Value != qualifierValue) && HoldTimeValid) {
                    // Change is valid ... 
                    //#region Value has changed
                    DeviceModel.parameters[0].Value = qualifierValue;
                    DeviceModel.parameters[0].TimeChanged = timeChanged;

                    // Update real-time value, including SNMP because of threshold change
                    this.sendEvent(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters[0]);
                    //this.sendSocket(DeviceModel.port, DeviceModel.uuid, "Value", DeviceModel.parameters[0]);

                    //Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Publish: UPDATE] Door=" + DeviceModel.parameters[0].Value + " (HoldTime=" + (DeviceModel.parameters[0].HoldTime == 0 ? "Not Used)" : parseFloat(HoldTimeChanged).toFixed(2) + "s)"));
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Publish: UPDATE] Door=" + JSON.stringify(DeviceModel.parameters[0]));

                    //#endregion Value has changed
                }
                //#endregion Publish logic
            }
        }// if (DeviceModel.parameters[0].Enabled === "YES")
        catch (error) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: Process Error=" + error);
        }
        finally {
            // Nothing here for now ...
        }
    }
    //#endregion processDigitalInput
    //#endregion Digital I/O 

    // Convert a raw value to a ranged value
    //#region convertRawValueToValue
    this.convertRawValueToValue = function (rawValue) {
        var rangedValue = Number(parseFloat(rawValue) * ANALOG_INPUTS_RAWVALUE_SLOPE + ANALOG_INPUTS_RAWVALUE_OFFSET);
        if (rangedValue < 0) rangedValue = 0;
        return rangedValue;
    }
    //#endregion convertRawValueToValue

    // Convert a ranged value to raw value
    //#region convertValueToRawValue
    this.convertValueToRawValue = function (rangedValue) {
        var rawValue = Math.round((parseFloat(rangedValue) - ANALOG_INPUTS_RAWVALUE_OFFSET) / ANALOG_INPUTS_RAWVALUE_SLOPE);
        return rawValue;
    }
    //#endregion convertValueToRawValue

    //#endregion Process I/O
    //#endregion Class Methods (Device Processing - Process I/O)
    //#endregion Class Methods (Device Processing)

    //#region Class Methods (Eventing)

    //#region Class Methods (Eventing - SendEvent)
    this.sendEvent = function (port, uuid, parameter) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".EVENT: Port=" + port + " Device='" + DeviceModel.title + "' Parameter='" + parameter.Id + "' Value='" + parameter.Value + "'");

        ThisDeviceObject.sendSocket(port, uuid, "Value", parameter)
        this.sendSMTP(port, uuid, parameter);
        this.sendSNMP(port, uuid, parameter);

        // EventLog
        //#region EventLog
        // Retrieve EventLog array
        if (fs.existsSync("log/port" + port + "events.json")) {
            fs.readFile("log/port" + port + "events.json", function (error, eventlog) {
                if (error) {
                    EventLog = [];// Error in EventLog retrieval
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog retrieve error. Error=" + error);
                } else {
                    EventLog = JSON.parse(eventlog);
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog retrieved. EventLog=" + JSON.stringify(EventLog));
                }

                // Update EventLog with latest event
                EventLog.unshift(parameter);// To the front of the array
                if (EventLog.length > EVENTS_LENGTH) {
                    EventLog.splice(-1);// Remove from the end if limit reached
                }

                // Save updated EventLog array          
                fs.truncate("log/port" + port + "events.json", 0, function () {
                    fs.writeFile("log/port" + port + "events.json", JSON.stringify(EventLog), function (error) {
                        if (error) {
                            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog save error. Error=" + error);
                        } else {
                            // Send updated EventLog
                            ThisDeviceObject.sendSocket(port, uuid, "EventLog", EventLog);// For real-time update

                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog updated and saved. EventLog=" + JSON.stringify(EventLog));
                            cp.exec("sync", function (error, stdout, stderr) {
                                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync stdout=" + stdout);
                                if (error) {
                                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync error=" + error);
                                }
                            });
                        }
                    });
                });

            });
        } else {
            // Update EventLog with latest event
            EventLog = [];// Empty event log
            EventLog.unshift(parameter);// To the front of the array
            

            // Save updated EventLog array
            fs.truncate("log/port" + port + "events.json", 0, function () {
                fs.writeFile("log/port" + port + "events.json", JSON.stringify(EventLog), function (error) {
                    if (error) {
                        Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog save error. Error=" + error);
                    } else {
                        // Send updated EventLog
                        ThisDeviceObject.sendSocket(port, uuid, "EventLog", EventLog);// For real-time update

                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog updated and saved. EventLog=" + JSON.stringify(EventLog));
                        cp.exec("sync", function (error, stdout, stderr) {
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync stdout=" + stdout);
                            if (error) {
                                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync error=" + error);
                            }
                        });
                    }
                });
            });
        }

        //#endregion EventLog
    }
    //#endregion Class Methods (Eventing - SendEvent)

    //#region Class Methods (Eventing - Real-time Updates)
    this.sendSocket = function (port, uuid, message, parameter) {
        //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSocket: Port=" + port + " Message=" + message + " Parameter=" + JSON.stringify(parameter));
        SocketIo.sockets.emit(DeviceModel.id, { "Port": port, "UUID": uuid, "Message": message, "Parameter": parameter })
    }
    //#endregion Class Methods (Eventing - Real-time Updates)

    //#region Class Methods (Eventing - SMTP)
    this.sendSMTP = function (port, uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (!SendEventsFlag) return;

        // Set up the SMTP Email message
        var portNumber = (parseInt(port) + 1);
        var event = "Status=" + (parameter.Value?"Open":"Closed") + " '" + parameter.Name + "' on '" + DeviceModel.title + "' (Port " + portNumber + ")";
        var ioInfo = "Peak: Port: " + portNumber + " '" + parameter.Name + "'";

        var message = "Timestamp=" + moment(parameter.TimeChanged).format() + "\n";
        message += "Port=" + portNumber + "\n";
        message += "Device=" + DeviceModel.title + "\n";
        message += "Parameter=" + parameter.Name + "\n";
        message += "Value=" + (parameter.Value?"Open":"Closed") + "\n";
        message += "Polarity=" + (parameter.Mode?"Normally Open":"Normally Closed") + "\n";
        message += "Hold Time=" + parseInt(parameter.HoldTime) + "\n";

        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: Sending SMTP Event for Port " + portNumber + " '" + parameter.Name + "'");
        SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
            Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: SMTP Email Event Response for Port " + portNumber + " '" + parameter.Id + "' " + result.result);
        })
    }
    //#endregion Class Methods (Eventing - SMTP)

    //#region Class Methods (Eventing - SNMP)
    this.sendSNMP = function (port, uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (!SendEventsFlag) return;

        // Set up the SNMP trap    
        var trapOid = "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".10.1.0.1";
        var varbinds = [
            {
                // Timestamp
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.1",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(moment().toString())
            },
            {
                // Port
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.2",
                type: NetSnmp.ObjectType.Integer,
                value: parseInt((parseInt(port) + 1))
            },
            {
                // Device
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.3",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(DeviceModel.title)
            },
            {
                // Device Parameter
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.4",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Name)
            },
            {
                // Device Parameter - Value
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.5",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer((parameter.Value?"Open":"Closed"))
            },
            {
                // Device Parameter - Mode
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.6",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer((parameter.Mode ? "Normally Open" : "Normally Closed"))
            },
            {
                // Device Parameter - HoldTime
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.7",
                type: NetSnmp.ObjectType.Integer,
                value: parseInt(parameter.HoldTime)
            }
        ];

        // Send the SNMP notification
        for (var i = 0; i < SnmpHostSessionsArray.length; i++) {
            // Skip if this host is not enabled
            if (SnmpHostSessionsArray[i].Enabled === "FALSE") continue;

            // Send trap to host ...
            var IPAddress = SnmpHostSessionsArray[i].Session.target;
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Sending SNMP trap to IPAddress=" + IPAddress);

            SnmpHostSessionsArray[i].Session.trap(trapOid, varbinds, function (error) {
                if (error) {
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Error=" + error);
                }
                else {
                    Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Sending to SNMP Host @ " + IPAddress + " - Port[" + varbinds[1].value + "] '" + varbinds[2].value + "' Value=" + varbinds[4].value);
                }
            });
        }

    }
    //#endregion Class Methods (Eventing - SNMP)

    //#endregion Class Methods (Eventing)

    //#endregion Class Methods
}

//#region Device Constructor
// Device constructor to ensure unique objects
//  Uses the following signature:
//      port        - port on which device is installed [0 indexed]
//      logger      - system logger object
//      io          - system socket io object
//      smtpManager - SMTP messaging object
//      netSnmp     - SNMP infrastructure object
module.exports = function (port, logger, io, smtpManager, netSnmp, callback) {
    var DeviceObject = new Device(port, logger, io, smtpManager, netSnmp);
    callback(DeviceObject);
}
//#endregion Device Constructor
//#endregion Device Class: Variables, Instantiation, Methods and Constructor

// Exported Methods
//#region Exported Methods

// Basic Info
//#region Basic Info
Device.prototype.SetPort = function (port) {
    this.setPort(port);
}

Device.prototype.GetPort = function () {
    return this.getPort()
}

Device.prototype.GetDeviceId = function () {
    return this.getDeviceId()
}

Device.prototype.GetDeviceModel = function () {
    return this.getDeviceModel()
}

Device.prototype.GetDeviceInformation = function (callback) {
    var deviceInfo = this.getDeviceInformation();
    callback(deviceInfo);
}

//#endregion Basic Info

// Enable/Disable
//#region Enable/Device
Device.prototype.Enable = function (enabled) {
    this.enable(enabled);
}
//#endregion Enable/Disable

// Send Events
//#region Send Events
Device.prototype.SendEvents = function (sendevents) {
    var id = this.getDeviceId();
    Logger.trace(id + ".SendEvents: SendEventsFlag=" + (sendevents ? "YES" : "NO"));
    this.sendEvents(sendevents);
}
//#endregion Send Events

// Update Device
//#region Update Device

// UpdateDevice
Device.prototype.UpdateDevice = function (port, parameter, model, callback) {
    var id = this.getDeviceId();
    this.updateDevice(port, parameter, model, function (model) {
        Logger.trace(id + ".UpdateDevice: Updated DeviceModel=" + JSON.stringify(model));
        callback(model);
    });
}

// SetSnmpHostSessionsArray
Device.prototype.SetSnmpHostSessionsArray = function (snmpHostSessions, callback) {
    var id = this.getDeviceId();
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    Logger.trace(id + ".SetSnmpHostSessionsArray: SNMP Host Sessions established");
    callback();
}

//#endregion Update Device

//#endregion ExportedMethods